import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class colonize_v4 extends PApplet {

// Still to do:
//  - Add buttons for zoom and to confirm picture choice/terrain changes
//  - Make it so you can't the camera too far out and lose the map


ArrayList people;       // A people thing
ArrayList workers;
ArrayList pioneers;     // Pioneeeersssees

Boat boat;

Slider seaLevelSlider;
Slider landHeightSlider;

FadingText instructions;

Tile[] tiles;           // An array of tile objects

int state = 0;          // State of program

int mouseTimer = 0;     // Timer to make sure the mouse only clicks once
boolean mouseClicked;   // Boolean that activates when mouse is clicked, deactivates after set ammount of millis

PImage sourceImg;       // Base image used to generate the map
PImage terrainMap;      // The map image to be generated
boolean invert;         // Will reverse land/water
boolean image_chosen;

//float avgBright;        // Average brightness of the image's pixels

float waterThreshold = 226;  // Threshold below which brightness of image becomes water
float landThreshold  = 253;  // Threshold below which brightness of image becomes land (above becomes mountains)

float s = 1;       // Scaling for errything
float transX = 0;  // Translation for everything
float transY = 0;  // Also that

int settlementThreshold;  // Number of people required to increase the settlement level of a tile.

Alarm personMoveTimer;  // A timer for 'turns'

float time; // For perlin noise


public void setup()
{
  size(800, 600);
  frameRate(30);
  background(128);
  frame.setResizable(true);
    
  instructions = new FadingText("Click to confirm!");
  
  noSmooth();
  
  loadPixels();
      
  selectInput("Choose an image to convert into terrain:", "fileSelected");
  
  settlementThreshold = 5;
  
  personMoveTimer = new Alarm(250);
  
  // Create an array list of People objects
  people = new ArrayList();
  
  workers = new ArrayList();
  
  // Create an array list of Pioneer objects
  pioneers = new ArrayList();
  
}


public void draw()
{
  loadPixels();
  
  noStroke();
  if (s > 1)
  {
    float tx = width/2 + transX*s;
    float ty = height/2 + transY*s;
//    ty = constrain(ty, -height*s, 0);
    translate(tx, ty);
    scale(s);
    translate(0, 0);
  } else {
    transX = -width/2;
    transY = -height/2;
  }

  // State 0: Taking the picture
  if (state == 0)
  {    
    if (image_chosen) {
      image(sourceImg, 0, 0);
      textSize(20);
      if (!mousePressed) text("click to confirm image!", width/2, height/2);
    } else {
      // Display choose image message
      fill( map(noise(time), 0, 1, 0, 255),
            map(noise(time+9383), 0, 1, 0, 255),
            map(noise(time-483286), 0, 1, 0, 255)     );
      textSize(  map( noise(time+338383), 0, 1, 5, 100) );
      textAlign(CENTER);
      text("Choose an image!", width/2, map( noise(time-2929200), 0, 1, 0, height) );
      fill(255);
      textSize(30);
      text("Welcome! This program will generate a terrain map from an image file of your choice! Your generated terrain will then be colonized by hordes of ship-bound pixels, eager to exploit all of your image's precious natural resources. They will found cities, explore, farm, mine and build roads. Then, they'll spawn offspring to do it all over again! Watch as your lush virgin wilderness is converted into an barely-habitable industrial hellscape, just like in real life! Note: performance drops significantly as image size increases.", 20, 20, width-20, height-20);
    }
  } // state 0 end (see 'cont_mouse' tab for next actions)


  // State 1: Converting the picture into the terrain map, and adjusting water-level / land height.
  if (state == 1)
  {
    // Create new image from frame when mouse was pressed
    sourceImg = loadImage("source.jpg");

    // Generate new terrain image from source image
    terrainMap = generateTerrain(sourceImg, waterThreshold, landThreshold);
    
    // Display the terrain map
    image(terrainMap, 0, 0);
    
    fill(255);
    textSize(10);
    text("Modify your terrain!", width/2, height*0.27f);
    text("Then, click anywhere to continue!", width/2, height*0.37f);
    
    // Display slider bars
    stroke(255);
    fill(105, 214, 87);
    seaLevelSlider.update();
    seaLevelSlider.display();
    fill(231, 178, 17);
    landHeightSlider.update();
    landHeightSlider.display();
    
    
    // Update thresholds
    waterThreshold = seaLevelSlider.output_value;
    landThreshold = landHeightSlider.output_value;
  } // state 1 end (see cont_Mouse for next actions)


  // State 2: Finalizing the map image and filling the array of Tile objects
  if (state == 2)
  { 
    analyzeTerrain();
    
    // Create boat
    boat = new Boat();
        
    println("State 3");
    state = 3;
  } // state 2 end
  
  
  // State 3: Where everything happens
  if (state == 3)
  {
    background(120);
    
    image(terrainMap, 0, 0);
    
    personMoveTimer.run();
    
    boat.move();
    boat.display();
    
    // Display the bottom layer
    for (int i = 0; i < pixels.length; i++) {
        tiles[i].belowPeopleDisplay();
    }  // for i = 0 end
    
    // Display the people layer
    if (people.size() > 0)
    {
      for (int i = 0; i < people.size(); i++)
      {
        Person p = (Person) people.get(i);
        if (p.disable == false) {
          if (personMoveTimer.finished) p.move();
          p.display();
        } else {
          p = null;
        }
      }
    }
    
    // Display the worker layer
    if (workers.size() > 0)
    {
      for (int i = 0; i < workers.size(); i++)
      {
        Worker w = (Worker) workers.get(i);
        if (w.disable == false) {
          if (personMoveTimer.finished) w.move();
          w.display();
        } else {
          w = null;
        }
      }
    }
    
    if (personMoveTimer.finished) personMoveTimer = new Alarm(100);
    
    // Display pioneers
    if (pioneers.size() > 0)
    {
      for (int i = 0; i < pioneers.size(); i++)
      {
        Pioneer p = (Pioneer) pioneers.get(i);
        p.timer();
        p.display();
        if (p.landed) p = null;
      }
    }
    
    // Display the top layer
    for (int i = 0; i < pixels.length; i++) {
        tiles[i].abovePeopleDisplay();
    }  // for i = 0 end
    
    fill(255);
    textSize(10);
    textAlign(LEFT);
    text("X: zoom in ~~~ Z: zoom out ~~~ Click: move camera", 10, height - 3);
    
  }  // state 3 end
  
  time += 0.005f;
  
} // end draw()
class Alarm
{
  float targetMillis;
  boolean finished;

  Alarm(float targetMillis_)
  {
    targetMillis = targetMillis_ + millis();
  }

  public void run()
  {
    if (millis() >= targetMillis) finished = true;
    else finished = false;
  }
}
class Boat
{
  int x, y;             // Position
  float velX, velY;     // Velocity
  
  boolean landed;       // Whether the ship has landed
  
  float randomValue;    // Holds a random value
  
  Boat()
  {
    landed = false;
    findStart();
    findRoute();
  }  // end of constructor
  
  // Find a route
  
  
  
  public void display()
  {
    fill(0,255,255);
    ellipse(x, y, 3, 3);
  }
  
  public void move()
  { 
    int loc = x + y * width;
    constrain(loc, 0, tiles.length-1);
    
    // If boat moves off-screen, get rid of it and spawn another boat
    if (x > width || y > height || x < 0 || y < 0)
    {
      if (landed == false)
      {
        boat = new Boat();
      }
    }
    
    // If boat reaches land, settle    
    if (landed == false && tiles[loc].terrainType == 1)
    {
      tiles[loc].terrainType = 4;
      tiles[loc].settlementLevel++;
      for(int i = 0; i < 13; i++)
      {
        PVector spawn = new PVector( x+random(velX*2),y+random(velY*2) );
        people.add(new Person( PApplet.parseInt(spawn.x), PApplet.parseInt(spawn.y), x, y) );
      }
      // Move boat off screen
      landed = true;
      x = -1000;
      y = -1000;
      velX = 0;
      velY = 0;
    }
    
    x += velX;
    y += velY;
  }
  
  public void findStart()
  {    
    // Find starting position
    randomValue = random(4);
    if (randomValue < 1)
    {
      x = PApplet.parseInt(random(0, width-1));
      y = 0;
    }
    else if (randomValue < 2)
    {
      x = width-1;
      y = PApplet.parseInt(random(0, height-1));
    }
    else if (randomValue < 3)
    {
      x = PApplet.parseInt(random(0, width-1));
      y = height-1;
    }
    else if (randomValue < 4)
    {
      x = 0;
      y = PApplet.parseInt(random(0, height-1));
    } 
        
    int loc = x + y * width-1;
        
    if (tiles[loc].terrainType == 0)  // If the selected tile is water...
    {
      return;
    }
    else
    {
      findStart();
    }
    
  }  // end of findStart()
  
  
  public void findRoute()
  {
    if (randomValue < 1)
    {
      velX = PApplet.parseInt(random(-4,  4));
      velY = PApplet.parseInt(random( 1,  3));
    }
    else if (randomValue < 2)
    {
      velX = PApplet.parseInt(random(-3, -1));
      velY = PApplet.parseInt(random(-4,  4));
    }
    else if (randomValue < 3)
    {
      velX = PApplet.parseInt(random(-4,  4));
      velY = PApplet.parseInt(random(-3, -1));
    }
    else if (randomValue < 4)
    {
      velX = PApplet.parseInt(random( 1,  3));
      velY = PApplet.parseInt(random(-4,  4));
    } 
  
    int tempX = x;
    int tempY = y;
    int landCounter = 0;
    
    for (int i = 0; i < height; i++)
    {
      tempX += velX;
      tempY += velY;
      
      int loc = tempX+tempY*width;
      loc = constrain(loc, 0, pixels.length-1);
      
      if (tiles[loc].terrainType == 1)
      {
        landCounter++;
      }
    }
    
    println(landCounter);
    
    if (landCounter >= 10)
    {
      return;
    }
    else
    {
      findRoute();
    }
  
  } // end of findRoute()


}  // end of Class

class Person
{
  int x, y;               // Position
  int col;              // Color

  int endurance;          // Modifier for moving on uneven terrain
  int sociability;        // Modifier for moving to a square with another person in it
  int terrainPreference;  // What type of terrain does this Person prefer?

  boolean disable;
  int homeX, homeY;       // Location of this Person's 'home' (where they first spawned)

  Person(int x_, int y_, int hx_, int hy_)
  {
    x = x_;
    y = y_;
    
    if (tiles[x+y*terrainMap.width].terrainType == 0 || tiles[x+y*terrainMap.width].terrainType == 2) disable = true;

    col = color(255, 200);

    endurance = PApplet.parseInt( random(0, 100) );  // If endurance is greater than or equal to the flatness of tile, there is a greater probability of moving there
    sociability = PApplet.parseInt( random(1, 5) );  // If sociability is greater than or equal to the number of people on a given tile, there is a greater probability of moving there
    terrainPreference = PApplet.parseInt( random(2) );  // 0 = fertility, 1 = mineral wealth (shouldn't ever actually equal 2)

    homeX = hx_;
    homeY = hy_;
  }


  public void move()
  {
    // Determine 1D location of current tile
    int loc = x + y*terrainMap.width;
  
    // Set this tile as having been used one more time
    tiles[loc].timesUsed++;
  
    // Call the current tile's 'settle' function. (ie: Check to see if a settlement should be created here.)
    if (tiles[loc].terrainType != 0) tiles[loc].settle();

    // 'Harvest' from current tile
    if (tiles[loc].brightness <= 80) tiles[loc].amountFarmed += tiles[loc].fertility;
    if (tiles[loc].brightness > 80)tiles[loc].amountMined  += tiles[loc].minWealth;
    if (tiles[loc].amountFarmed >= tiles[loc].farmThreshold && tiles[loc].terrainType == 1)  // If this tile is turned into a farm...
    {
      // Create a new farmer
      workers.add( new Worker(x, y, homeX, homeY, x, y, 0) );
      // Disable this person
      disable = true;
      tiles[loc].terrainType = 3;  // ...set it's terrain type to 3 (for farm)...
      return;
    }
    if (tiles[loc].amountMined >= tiles[loc].mineThreshold && tiles[loc].terrainType == 1)  // If this tile is turned into a mine...
    {
      // Create a new farmer
      workers.add( new Worker(x, y, homeX, homeY, x, y, 1) );
      // Disable this person
      disable = true;
      tiles[loc].terrainType = 5;  // ...set it's terrain type to 5 (for mine)...
      return;
    }

    // Create movement probability matrix and fill each entry with baseProb
    int[][] moveProb = new int[3][3];
    for (int i = 0; i < 3; i++) {
      for (int j = 0; j < 3; j++) {
        moveProb[i][j] = 0;
      }
    }

    float totalProb = 0;  // The added total of all 9 probability values

    // Check out all potential tiles and determine probability of moving to each one
    for (int i = 0; i < 3; i++)
    {
      for (int j = 0; j < 3; j++)
      {
        int tileLoc = (x + (i-1)) + (y + (j-1))*terrainMap.width;  // 1D location of the tile being looked at
        tileLoc = constrain(tileLoc, 0, tiles.length-1);
                  
        // Dealing with specific types of terrain
        if (tiles[tileLoc].terrainType == 1)
        {  // If the tile being looked at is land
        
          // Land tiles start with 1 probability always
          moveProb[i][j] = 1;
  
          // If endurance is greater than flatness, add 1 to this tile's probability
          if (endurance >= tiles[tileLoc].flatness) moveProb[i][j]++;
  
          // Deal with terrain preference
          if (terrainPreference == 0)  // If this Person prefers fertility...
          {
            if (tiles[i].fertility >= tiles[tileLoc].minWealth) moveProb[i][j]++;
          } else if (terrainPreference == 1)  // ... or else, if this Person prefers mineral wealth...
          {
            if (tiles[i].minWealth >= tiles[tileLoc].fertility) moveProb[i][j]++;
          }
        }

        // See how many people are in this tile
        int peopleInTile = 0;
        for (int k = 0; k < people.size (); k++)
        {
          Person p = (Person) people.get(k);
          if ((p.x + p.y * width) == tileLoc)
          {
            peopleInTile++;
          }
        }

        // If this person's sociability rating is greater than or equal to the number of people in that tile, increase the probability of moving there.
        if (sociability <= peopleInTile) moveProb[i][j]++;

        // Add current probability to total
        totalProb += moveProb[i][j];

      } // end for j
    }  // end for i

    // Generate a random number from 0 to totalProb
    float r = random(0, totalProb);
    float lowerRange = 0;
    float upperRange = 0;

    for (int i = 0; i < 3; i++)
    {
      for (int j = 0; j < 3; j++)
      {
        upperRange += moveProb[i][j];
        if (r > lowerRange && r <= upperRange)  // If movement is less than current moveProb
        {
          x += (i-1);
          x = constrain(x, 0, width-1);
          y += (j-1);
          y = constrain(y, 0, height-1);
          return;
        }
        else lowerRange = upperRange;
      }
    }

  } //<>//


  public void display()
  {
    fill(col);
    rect(x, y, 1, 1);
  }  // display() end
}  // end class
class Pioneer
{
  int x, y;             // Position
  int velX, velY;     // Velocity
  
  boolean landed;       // Whether the ship has landed
  
  float randomValue;    // Holds a random value
  
  int savedTime;           // Interval at which this Person moves (milliseconds)
  int totalTime;        // Current time in milliseconds
  
  Pioneer(int _x, int _y)
  {
    x = _x;
    y = _y;
    velX = PApplet.parseInt(random(-4,4));
    velY = PApplet.parseInt(random(-4,4));
    
    landed = false;
    
    totalTime = 250; 
    savedTime = millis();
  }  // end of constructor
  
  
  public void timer()
  {
    int passedTime = millis() - savedTime;
    if (passedTime > totalTime)
    {
      move();
      savedTime = millis();
    }  // timer end
  }  // move() end  
  
  
  public void display()
  {
    fill(255,255,0);
    rect(x, y, 1, 1);
  }
  
  public void move()
  {
    int loc = x + y * width;
    int nextLoc = (x+velX) + (y+velY)*width;
    loc = constrain(loc, 0, tiles.length-1);
    nextLoc = constrain(nextLoc, 0, tiles.length-1);
    tiles[loc].timesUsed += 20;
    
    if (landed == false)
    { 
      // If pioneer reaches a good spot, settle
      if (tiles[loc].mountainAdj == true || tiles[loc].waterAdj == true || tiles[nextLoc].terrainType == 4 || tiles[nextLoc].terrainType == 2 || tiles[nextLoc].terrainType == 0)
      {
        tiles[loc].terrainType = 4;
        tiles[loc].settlementLevel++;
        for(int i = 0; i < 3; i++)
        {
          PVector spawn = new PVector( x-random(velX*2),y-random(velY*2) );
          people.add(  new Person( PApplet.parseInt(spawn.x),PApplet.parseInt(spawn.y), x, y )  );
        }
        // Move pioneer off screen
        x = -1000;
        y = -1000;
        velX = 0;
        velY = 0;
        landed = true;
      }
      x += velX;
      y += velY;
    }
  }  // end of move()


}  // end of Class
class Tile
{
  int x, y;         // Position

  int terrainType;    // 0 = water, 1 = land, 2 = mountain, 3 = farm, 4 = settlement, 5 = mines
  float flatness;    // How level the current tile is compared to it's surroundings
  float timesUsed;      // Number of times a Person has been on this tile
  
  int col;           // Color of the tile
  float brightness;    // Average of rgb (used to calculate various other values; see below)

  // For water
  boolean landAdj;     // Is this pixel adjacent to land?
  float depth;         // Affects fishing potential & movement speed (based on brightness)

  // For land
  boolean waterAdj;       // Is this tile adjacent to a water tile?
  boolean mountainAdj;    // Is this tile adjacent to a mountain?
  
  float minWealth;          // Potential for mining (higher brightness = more minWealth)
  int amountMined;          // Total amount that has been mined from this tile.
  int mineThreshold;        // Amount of minerals necessary to be gathered before this tile becomes a mine
  
  int settlementLevel;      // Size of settlement upon tile.
  int storedFood;           // Amount of food stored in this settlement
  int storedMinerals;       // Amount of minerals stored in this settlement
  int spawnThreshold;       // Amount of food needed to be stored before this settlement spawns a new Person.
  int pioneerThreshold;     // Amount of minerals needed to be stored before thie settlement sends out a company of pioneers (by ship or otherwise)
  
  float fertility;          // Potential for agriculture (lower brightness = more fertility)
  int amountFarmed;         // Amount of food that has been gathered from this tile
  int farmThreshold;        // Amount of food necessary to be gathered before this tile becomes a farm

  Tile(int x_, int y_, int terrainType_, int col_)  // add the additional variables once this one is working.
  {
    x = x_;
    y = y_;
    terrainType = terrainType_; 
    col = col_;
    
    brightness = (red(col) + green(col) + blue(col)) / 3;
    
    // If land type is water, establish depth based on brightness
    if (terrainType == 0) {
      depth = map(brightness, 0, waterThreshold, 0, 100);
    }
    
    // If terrain type is land, establish fertility and mineral wealth based on brightness
    if (terrainType == 1) {
      fertility = map(brightness, 62, 98, 100,     0);
      minWealth = map(brightness, 62, 98,   0,   100);
    }
    
    map(flatness, 0, 100, -50, 50);
    // Modify fertility based on levelness/water adjacency (more levelness = more fertility)
    fertility += flatness;
    if (waterAdj == true) fertility += 15;
    fertility = map(fertility, -50, 165, 0, 100);
    
    timesUsed = 0;
    
    // Modify mineral wealth based on levelness/mountain adjancency (less levelness = more mineral wealth)
    minWealth -= flatness;
    if (mountainAdj == true) minWealth += 15;
    minWealth = map(minWealth, -50, 165, 0, 100);
    
    settlementLevel = 0;
    storedFood = 0;
    spawnThreshold = 1000;
    pioneerThreshold = 2000;
    
    amountMined = 0;
    mineThreshold = 650;
    
    amountFarmed = 0;
    farmThreshold = 650;
  }
  
  // Display tiles which should be below people layer
  public void belowPeopleDisplay()
  {
    // For coasts
    if (landAdj == true)
    {
      fill(151, 180, 238);
      rect(x, y, 2, 2);
    }
    // For roads
    if (timesUsed > 0)
    {
      fill(150, map(timesUsed, 0, 40, 0, 200));
      rect(x, y, 2, 2);
    }
    // For farms
    if (terrainType == 3)  
    {
      fill(255, 190, 0, 128);
      rect(x, y, 2, 2);
    }
    // For mines
    if (terrainType == 5)
    {
      fill(0, 128);
      rect(x, y, 2, 2);
    }
  }  // display() end
  
  
  // Display tiles which should be above people layer
  public void abovePeopleDisplay()
  {
    // For settlements
    if (settlementLevel > 0)  
    {
      fill(map(storedMinerals, 0, pioneerThreshold, 120, 170), 120, map(storedFood, 0, spawnThreshold, 120, 170));
      rect(x, y, 2, 2);
    }

  }
  
  
  public void settle()
  {
    IntList storedPersons = new IntList();  // An integer list to store the array locations of Person objects
    // Check how many People are on this tile
    for (int i = 0; i < people.size(); i++)
      {
        Person p = (Person) people.get(i);
        int tloc =   x +   y*terrainMap.width;  // 1D location of this tile
        int ploc = p.x + p.y*terrainMap.width;  // 1D location of the Person being looked at
        if (ploc == tloc)  // If the person being looked at is on this tile...
        {
          storedPersons.append(i);  // ...store the index number of this Person.
        }
      }
      if (storedPersons.size() >= settlementThreshold && settlementLevel == 0)  // If the number of people on this tile is at the settlement threshold...
      {
        println("Settlement increase!");
        // ...and delete a number of people equal to the settlement threshold...
        storedPersons.sortReverse();
        for (int j = 0; j < settlementThreshold; j++)
        {
          int sp = storedPersons.get(j);
          people.remove(sp);
        }
        // Then create a new person at this settlement
        people.add(new Person(x, y, x, y));
        terrainType = 4;    // (settlements have their own terrainType)
        settlementLevel++;  // ...increase the settlement level...
      }
      storedPersons.clear();  // Clear the list of stored people
  }
  
  
  public void spawn()
  {
      people.add(new Person(x, y, x, y));
      println("Person was born! Woop there it is!");
      storedFood -= spawnThreshold;
  } // end spawn()
  
  
  public void pioneer()
  {
    // Create a new pioneer
    pioneers.add(new Pioneer(x, y));
    storedMinerals -= pioneerThreshold;
    pioneerThreshold += 2000;
  }
  
} // end class
class Worker
{
  int x, y;               // Position
  int col;              // Color

  int jobType;            // 0 = farmer, 1 = miner;
  boolean disable;
  boolean returnHome;     // Whether this guy is returning home from his job or vica versa
  int homeX, homeY;       // Location of this Person's 'home' (where they first spawned)
  int jobX, jobY;         // Location of this Person's 'job'

  int food;               // Amount of food held by this character
  int minerals;           // Amount of minerals held by the character

  boolean foundPath;

  Worker(int x_, int y_, int hx_, int hy_, int jx_, int jy_, int jobType_)
  {
    x = x_;
    y = y_;
    
    jobType = jobType_;
    
    if (tiles[x+y*terrainMap.width].terrainType == 0 || tiles[x+y*terrainMap.width].terrainType == 2) disable = true;

    col = color(255, 200);

    returnHome = true;
    foundPath = true;
    homeX = hx_;
    homeY = hy_;
    jobX = jx_;
    jobY = jy_;

    food = 0;
    minerals = 0;
  }


  public void move()
  {
    // Determine 1D location of current tile
    int loc = x + y*terrainMap.width;

    // Set this tile as having been used one more time
    tiles[loc].timesUsed++;

    // Call the current tile's 'settle' function. (ie: Check to see if a settlement should be created here.)
    if (tiles[loc].terrainType != 0) tiles[loc].settle();


    // If this Person's jobType is a farmer:
    if (jobType == 0)
    {

      if (x == jobX && y == jobY)  // If this person is at their job location...
      {
        food += tiles[loc].fertility;  // ...gather resources...
        returnHome = true;  // ...and return home.
      }

      if (x == homeX && y == homeY)  // If this person is at their home...
      {
        tiles[loc].storedFood += food;  // ...add their carried resources to their home's stored food.
        food = 0;   
        if (tiles[loc].storedFood >= tiles[loc].spawnThreshold) tiles[loc].spawn();  // Spawn a new Person if necessary
        returnHome = false;  // ...and return to their job location.
//        if (tiles[loc].settlementLevel == 0)
//        {
//          jobType = 0;
//        }
      }

      if      (returnHome == true)  goTo(homeX, homeY);
      else if (returnHome == false) goTo( jobX, jobY);
    }  // jobType = 1 end

    // If this Person's jobType is a miner:
    else if (jobType == 1)
    {

      if (x == jobX && y == jobY)  // If this person is at their job location...
      {
        minerals += tiles[loc].minWealth;  // ...gather resources...
        returnHome = true;  // ...and return home.
      }

      if (x == homeX && y == homeY)  // If this person is at their home...
      {
        tiles[loc].storedMinerals += minerals;  // ...add their carried resources to their home's stored minerals.
        if (tiles[loc].storedMinerals >= tiles[loc].pioneerThreshold) tiles[loc].pioneer();  // Spawn a new Pioneer if necessary
        minerals = 0;   
        returnHome = false;  // ...and return to their job location.
//        if (tiles[loc].settlementLevel == 0)
//        {
//          jobType = 0;
//        }
      }

      if      (returnHome == true)  goTo(homeX, homeY);
      else if (returnHome == false) goTo( jobX, jobY);
    }  // jobType = 2 end
  }  // move() end


  // Function to move towards a certain point
  public void goTo(int destX, int destY)
  {
    // Create movement probability matrix and fill each entry with 0
    int[][] moveProb = new int[3][3];
    for (int i = 0; i < 3; i++)
    {
      for (int j = 0; j < 3; j++)
      {
        int tileLoc = (x + (i-1)) + (y + (j-1))*terrainMap.width;  // 1D location of the tile being looked at
        
        if (tiles[tileLoc].terrainType != 0 && tiles[tileLoc].terrainType != 2)
        {  // If the tile has been moved on before
          moveProb[i][j] = 1;  // ...It can be moved to.
        } else moveProb[i][j] = 0;
      }
    }
    
    // Find a possible target location
    int possibleTargetX;
    int possibleTargetY;
    
    // If the last path was successful
    if (foundPath == true) {
      if (destX > x) possibleTargetX = 1;
      else if (destX < x) possibleTargetX = -1;
      else possibleTargetX = 0;
      
      if (destY > y) possibleTargetY = 1;
      else if (destY < y) possibleTargetY = -1;
      else possibleTargetY = 0;
    } else {
      // Get a new target location
      possibleTargetX = PApplet.parseInt( random(-2, 2) );
      possibleTargetY = PApplet.parseInt( random(-2, 2) );
    }
    
    // Test path
    if ( moveProb[possibleTargetX+1][possibleTargetY+1] == 0 ) foundPath = false;
    else foundPath = true;
    
    if (foundPath == true) {
      x += possibleTargetX;
      y += possibleTargetY;
    }
  }  // end of function


  public void display()
  {
    fill(col);
    rect(x, y, 1, 1);
  }  // display() end
}  // end class
class FadingText {
  PVector location;
  String text;
  float alpha;
  
  FadingText(String text_) {
    text = text_;
    alpha = 255;
    location = new PVector(0, 0);
  }
  
  public void display(float x, float y) {
    location = new PVector(x, y);
    fill(255, alpha);
    textAlign(CENTER);
    text(text, location.x, location.y);
    alpha -= 1;
  }
}
class Slider {
  PVector location, slider_location;
  float lowest_value, highest_value, current_value;
  float output_value;
  float size;
  String label;
  boolean  mouse_over;
  
  Slider(float current_value_, float lowest_value_, float highest_value_, float x_, float y_, float size_, String label_) {
    location = new PVector(x_, y_);
    lowest_value = lowest_value_;
    highest_value = highest_value_;
    current_value = current_value_;
    size = size_;
    label = label_;
    slider_location = new PVector( map(current_value, lowest_value, highest_value, location.x-size/2, location.x+size/2),
                                   location.y );
  }
  
  
  public void update() {
    // See if mouse is over this slider
    if (mouseX >= slider_location.x-5 && mouseX <= slider_location.x+5 && mouseY >= slider_location.y-10 && mouseY <= slider_location.y+10) mouse_over = true;
    else mouse_over = false;
    
    // Constrain the slider to stay within the bar
    slider_location.x = constrain(slider_location.x, location.x-size/2, location.x+size/2);
    
    output_value = map( slider_location.x, location.x - size/2, location.x+size/2, lowest_value, highest_value);
  }
  
  
  public void display() {
    // Display bar
    line(location.x-size/2, location.y, location.x+size/2, location.y);
    line(location.x-size/2, location.y+10, location.x-size/2, location.y-10);
    line(location.x+size/2, location.y+10, location.x+size/2, location.y-10);
    
    // Display slider
    rectMode(CENTER);
    rect(slider_location.x, slider_location.y, 10, 20);
    rectMode(CORNER);
    
    fill(255);
    textSize(10);
    text(label, location.x, location.y+10);
  }
  
}
public void keyPressed()
{
  if (state == 1)
  {
//    if (key == CODED) {
//      if (keyCode == UP) {
//        waterThreshold ++;
//      } // UP end
//      if (keyCode == DOWN) {
//        waterThreshold --;
//      } // DOWN end
//      if (keyCode == RIGHT) {
//        landThreshold ++;
//      } // RIGHT end
//      if (keyCode == LEFT) {
//        landThreshold --;
//      } // LEFT end
//    } // if key == CODED end
//    
//    if (key == ' ') {
//        invert = !invert;
//      }
      
      waterThreshold = constrain(waterThreshold, 0, landThreshold);
      landThreshold = constrain(landThreshold, waterThreshold, 255);
      
      println("Current Water Threshold: " + waterThreshold);
      println("Current Land Threshold: " + landThreshold);
  } // if (state == 1) end
  
  if (state == 3)
  {
    if (key == 's')
    {
    // Create new Settlement
    int mloc = mouseX + mouseY * width;
    tiles[mloc].terrainType = 4;
    tiles[mloc].settlementLevel++;
    for(int i = 0; i < 11; i++)
    {
      people.add(new Person((mouseX+PApplet.parseInt(random(-2,2))),(mouseY+PApplet.parseInt(random(-2,2))), mouseX, mouseY));
    }
    println("fertility:" +tiles[mloc].fertility);
    println("brightness:" +tiles[mloc].brightness);
    }
    
    if (key == 'p')
    {
      // Create a new pioneer
      pioneers.add(new Pioneer(mouseX, mouseY));
    }
    
    if (key == 'z') s -= 1;
    if (key == 'x') s += 1;
    s = constrain(s, 1,10);
    println(s);
    
  }  // end state == 3
        
} // end keyPressed()
public void mouseClicked()
{

  if (state == 0)
  {    
    // Convert picture to grayscale and invert the colors
    filter(GRAY);
    filter(INVERT);
    // Save an image
    saveFrame("source.jpg");
    
    // Create slider bars
    seaLevelSlider = new Slider(waterThreshold, 0, 255, width*0.25f, height*0.8f, width*0.35f, "sea level");
    landHeightSlider = new Slider(landThreshold, 255, 200, width*0.75f, height*0.8f, width*0.35f, "land height");
    
    // Create an array of Tile objects the same size as the window
    loadPixels();
    tiles = new Tile[pixels.length/2];
        
    // Advance to state 1
    state = 1;
    println("State 1");
    mouseClicked = true;
  } // end if state == 0


  if (state == 1 && mouseClicked == false)
  {    
    // Advance to state 2
    println("State 2");
    state = 2;
    mouseClicked = true;
  }  // end if (state == 1)
  
  if (state == 3 && mouseClicked == false)
  {
    transX += (width/2 - mouseX)/s;
    transY += (height/2 - mouseY)/s;
    mouseClicked = true;
  }  // end state 3
  
  
}  // end mouseClicked()


public void mouseReleased()
{
  mouseClicked = false;
} // end mouseReleased()



public void mouseDragged() {
  if (state == 1) {
    if (seaLevelSlider.mouse_over) seaLevelSlider.slider_location.x += mouseX - seaLevelSlider.slider_location.x;
    if (landHeightSlider.mouse_over) landHeightSlider.slider_location.x += mouseX - landHeightSlider.slider_location.x;
  }
}
// Run through tiles array and check the tile's proximity to other types of tiles...
// then apply the correct values to it.
public void analyzeTerrain()
{
  for (int x = 0; x < terrainMap.width/2; x++) {
    for (int y = 0; y < terrainMap.height/2; y++) {  // Use terrainMap width and height as proxy for imaginary "image" of tiles
      
        // Get 1D location of current tile:
        int loc = x + y*(terrainMap.width/2);
        loc = constrain(loc, 0, tiles.length-1);
    
        // Determine this tile's flatness rating
        float flatness = 0;
        // Here's an array to store the flatness ratings compared to each surrounding tile
        float[] flatnessRatings = new float[8];
        // And an index for that array:
        int flatnessRatingsIndex = 0;
        for (int v = 0; v < 3; v++) {
          for (int w = 0; w < 3; w++) {

            // Get the 1D location of the tile we're comparing looked at:
            int compareTileLoc = (tiles[loc].x + (v-1)) + (tiles[loc].y + (w-1))*terrainMap.width;
            compareTileLoc = constrain(compareTileLoc, 0, tiles.length-1);

            // Deal with flatness by comparing this tile's height to all of it's surrounding tiles:
            // Make sure not to count the current tile:
            if (compareTileLoc != loc) {
              flatnessRatings[flatnessRatingsIndex] = determineFlatness(loc, compareTileLoc);
              flatnessRatingsIndex++;
            }

            // Get the average of the flatness ratings for each surrounding tile:
            for (int q = 0; q < flatnessRatings.length; q++) { 
              flatness += flatnessRatings[q];
            } 
            flatness /= (float) flatnessRatings.length;
            tiles[loc].flatness = flatness;
          }
        }
      
    }  // end for y
  }  // end for x
  
}  // end of function
public void compareTiles(int thisLoc, int thatLoc)
{

  if (tiles[thisLoc].terrainType == 0)  // If this tile is water...
  {  // B
    if (tiles[thatLoc].terrainType != 0)  // ... and if that tile is not water...
    {  // C  
      tiles[thisLoc].landAdj = true;  // ...set this tile as land adjacent.
    }  // end C
  }  // end B
  else if (tiles[thisLoc].terrainType == 1)  // If this tile's is land...
  {  // B cont...
    
    if (tiles[thatLoc].terrainType == 0)  // ...and if that tile is water...
    {  // C
      tiles[thisLoc].waterAdj = true;  // ... set this tile as water adjacent.
    }  // end C
    else if (tiles[thatLoc].terrainType == 2)  // Or else if that tile is a mountain...
    {  // C cont...
      tiles[thisLoc].mountainAdj = true;  // ...set this tile as mountain adjacent.
    }  // end C
    else if (tiles[thatLoc].terrainType == 1)  // Or else if that tile is also land...
    {  // C cont...
                       // Space left here just in case
    }  // end C
  }  // end B
  
}  // end of function
public float determineFlatness(int thisLoc, int thatLoc)
{
  // Flatness is the difference between the brightness of the two tiles mapped to 0-100
  float flatness = abs(map(tiles[thisLoc].brightness, waterThreshold, landThreshold, 0, 100) - map(tiles[thatLoc].brightness, waterThreshold, landThreshold, 0, 100));
  return flatness;
}  // end of function
public void fileSelected(File selection) {
  if (selection == null) {
    println("Window was closed or the user hit cancel.");
  } else {
    sourceImg = loadImage(selection.getAbsolutePath());
    image_chosen = true;
    frame.setSize(sourceImg.width, sourceImg.height);
  }
}
public PImage generateTerrain(PImage source, float waterThreshold, float landThreshold)
{
  PImage destination;  // The image which this function will create
  destination = createImage(source.width, source.height, RGB);
  
  int terrainType = 0;

  if (invert)
  {
    source.filter(INVERT);
  }

  source.loadPixels();
  destination.loadPixels();

  // Going through the image every other pixel
  for (int x = 0; x < source.width-1; x+=2)
  {
    for (int y = 0; y < source.height-1; y+=2)
    {     
      
      // Find 1D pixel location
      int loc = x + y*source.width;

      // Find rgb values of current pixel
      float r = red  (source.pixels[loc]);
      float g = green(source.pixels[loc]);
      float b = blue (source.pixels[loc]);

      // Find the average of the current pixel's rgb
      float currentAvg = (r + g + b) / 3;

      //currentAvg = map(currentAvg, 0, 255, 0, avgBright);

      // Compare the brightness of this pixel to the brightness of the total image
      // and apply a color accordingly
      int water_distance = 15;
      if (currentAvg < waterThreshold || (x <= water_distance || x >= width-water_distance) || (y <= water_distance || y >= height-water_distance) )
      { // If brightness is below 1/3 of the average (or it is within a certain number of tiles of the edge of the window), give it water colors
        r = map(currentAvg, 0, waterThreshold, 65, 76);
        g = map(currentAvg, 0, waterThreshold, 70, 95);
        b = map(currentAvg, 0, waterThreshold, 210, 213);
        destination.pixels[loc] = color(r, g, b);

        // Also set its terrain type to 'water'
        terrainType = 0;  // 0 = water
      }
      else if (currentAvg < landThreshold)
      { // If brightness is below 2/3 of the average, set its color to green
        r = map(currentAvg, waterThreshold, landThreshold, 21, 130);
        g = map(currentAvg, waterThreshold, landThreshold, 122, 154);
        b = map(currentAvg, waterThreshold, landThreshold, 42, 8);
        destination.pixels[loc] = color(r, g, b);

        // Also set its terrain type to 'land'
        terrainType = 1;  // 1 = land
      }
      else
      { // If brightness is above 2/3 of the average, sit its color to red
        r = map(currentAvg, landThreshold, 255, 152, 182);
        g = map(currentAvg, landThreshold, 255, 113, 181);
        b = map(currentAvg, landThreshold, 255,  51, 179);
        destination.pixels[loc] = color(r, g, b);

        // Also set its terrain type to 'mountains'
        terrainType = 2;  // 2 = mountains
      }  // end if chain
      
      // Create a new Tile with corresponding location and terrain type
      tiles[loc] = new Tile(x, y, terrainType, color(r,g,b));
      
    }  // end for y
  }  // end for x

  destination.updatePixels();
  return destination;
} // end function
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "colonize_v4" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
